///////////////////////////////////////////////////////////////////////
// AC-3 Plugin for VirtualDub 1.8.1+
// Copyright (C) 2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Windows file I/O functions, wrapped in a class
// Files larger than 4GB are supported on NTFS.

#include <windows.h>

#include "infile.h"

#define INPAGE(x)		(x & (~(BUF_SIZE-1)))
#define INOFFSET(x)		(x & (BUF_SIZE-1))


static inline LPSTR wchar_to_ansi(LPCWSTR szWchar, int len) {
	LPSTR tmp = NULL;
	if (szWchar != NULL) {
		int lenA = WideCharToMultiByte(CP_ACP, 0, szWchar, len, NULL, 0, NULL, NULL);
		if (lenA > 0) {
			tmp = (LPSTR)malloc(lenA);
			if (WideCharToMultiByte(CP_ACP, 0, szWchar, len, tmp, lenA, NULL, NULL) != lenA) {
				free(tmp);
				tmp = NULL;
			}
		}
	}
	return tmp;
}


inFile::inFile() {
	hFile = INVALID_HANDLE_VALUE;
	fSize.i64 = 0;
	fPos.i64 = 0;
}

inFile::~inFile() {
	if (hFile != INVALID_HANDLE_VALUE) CloseHandle(hFile);
}

/*
	How Windows handles some aberrant situations:

	Setting the file pointer to absolute -10 returns 0xFFFFFFFF
	and error code 87 (ERROR_INVALID_PARAMETER), and does
	not move the file pointer (apparently)

	Setting the file pointer to relative -10 returns 0xFFFFFFFF
	and error code 131 (ERROR_NEGATIVE_SEEK), and does
	not move the file pointer

	Setting the file pointer AT or BEYOND the end of file succeeds.

	Attempting to read past the end of file does not cause
	an error, but reads zero bytes.
*/

void inFile::inRefresh() {
	DWORD dwRead, dwLo;

	dwRead = (DWORD)fPos.lHi;
	dwLo = INPAGE(fPos.dwLo);

	SetFilePointer(hFile, dwLo, (LONG *)&dwRead, FILE_BEGIN);

	if (ReadFile(hFile, bBuff, BUF_SIZE, &dwRead, NULL) == 0) {
		dwRead = 0;
	}

	if (dwRead < BUF_SIZE) {
		memset(bBuff + dwRead, 0, BUF_SIZE - dwRead);
	}
}

HANDLE inFile::dualCreateFile(LPCWSTR lpFileName, DWORD dwDesiredAccess,
	DWORD dwShareMode, LPSECURITY_ATTRIBUTES lpSecurityAttributes,
	DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes,
	HANDLE hTemplateFile)
{
	HANDLE h;

	if (GetVersion() < 0x80000000) {
		return CreateFileW(lpFileName, dwDesiredAccess, dwShareMode,
			lpSecurityAttributes, dwCreationDisposition,
			dwFlagsAndAttributes, hTemplateFile);
	}

	LPSTR tmp = wchar_to_ansi(lpFileName, -1);

	if (tmp != NULL) {
		h = CreateFileA(tmp, dwDesiredAccess, dwShareMode,
			lpSecurityAttributes, dwCreationDisposition,
			dwFlagsAndAttributes, hTemplateFile);

		free(tmp);
		return h;
	}

	return INVALID_HANDLE_VALUE;
}

bool inFile::inOpen(const wchar_t *szFile) {
	DWORD dwLo, dwHi;
	HANDLE h;

	bool ret = false;

	if (szFile == NULL) goto Abort;
	if (*szFile == 0) goto Abort;

	h = dualCreateFile((LPWSTR)szFile, GENERIC_READ, FILE_SHARE_READ, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS, NULL);

	if (h == INVALID_HANDLE_VALUE) goto Abort;

	dwLo = GetFileSize(h, &dwHi);
	if (dwLo == 0xFFFFFFFF) {
		if (GetLastError() != NO_ERROR) goto Abort2;
	}
	if (dwHi & 0x80000000) goto Abort2;

	if (hFile != INVALID_HANDLE_VALUE) CloseHandle(hFile);

	hFile = h;
	fPos.i64 = 0;
	fSize.dwLo = dwLo;
	fSize.lHi = (LONG)dwHi;

	inRefresh();
	ret = true;

Abort:
    return ret;

Abort2:
	CloseHandle(h);
	goto Abort;
}

__int64 inFile::inSize() {
	if (hFile != INVALID_HANDLE_VALUE) return fSize.i64;
	return 0;
}

__int64 inFile::inPos() {
	if (hFile != INVALID_HANDLE_VALUE) return fPos.i64;
	return 0;
}

bool inFile::inEOF() {
	if (hFile != INVALID_HANDLE_VALUE) return (fPos.i64 >= fSize.i64);
	return true;
}

bool inFile::inSeek(__int64 iAbsPos) {
	bool ret = false;

	if ((hFile != INVALID_HANDLE_VALUE) && iAbsPos >= 0) {

		ret = (INPAGE(iAbsPos) != INPAGE(fPos.i64));
		fPos.i64 = iAbsPos;

		if (ret) inRefresh();
		ret = true;
	}

	return ret;
}

long inFile::inRead(void *lpBuffer, long lBytes) {
	int remains, offs;
	long ret = 0;

	if (hFile == INVALID_HANDLE_VALUE) goto Abort;
	if (lpBuffer == NULL) goto Abort;
	if (lBytes <= 0) goto Abort;

	// Don't read past EOF
	{
		__int64 i = fSize.i64 - fPos.i64;
		if (i <= 0) goto Abort;
		if ((__int64)lBytes > i) lBytes = (long)i;
	}

	offs = INOFFSET(fPos.dwLo);
	remains = BUF_SIZE - offs;

	for (;;) {
		if (remains > lBytes) remains = lBytes;
		memcpy(lpBuffer, bBuff + offs, remains);
		ret += remains;
		fPos.i64 += remains;
		lBytes -= remains;
		if (lBytes <= 0) break;
		lpBuffer = (char *)lpBuffer + remains;
		inRefresh();
		offs = 0;
		remains = BUF_SIZE;
	}

	if (INOFFSET(fPos.dwLo) == 0) inRefresh();

Abort:
	return ret;
}

int inFile::inByte() {
	int ret = 0;

	if (hFile != INVALID_HANDLE_VALUE) {
		if (fPos.i64 < fSize.i64) {
			int offs = INOFFSET(fPos.dwLo);
			ret = bBuff[offs];
			++fPos.i64;
			if (offs == (BUF_SIZE - 1)) inRefresh();
		}
	}

	return ret;
}
