///////////////////////////////////////////////////////////////////////
// AC-3 Plugin for VirtualDub 1.8.1+
// Copyright (C) 2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>

#include "vdinputdriver.h"
#include "vdunknown.h"

#include "infile.h"
#include "resource.h"


static HMODULE g_hModule;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved) {
	if (fdwReason == DLL_PROCESS_ATTACH) {
		g_hModule = (HMODULE)hinstDLL;
	}
	return TRUE;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileAC3
//
///////////////////////////////////////////////////////////////////////

class InputFileAC3 : public vdxunknown<IVDXInputFile> {

public:
	InputFileAC3(const VDInputDriverContext& context);

	void VDXAPIENTRY Init(const wchar_t *szFile, IVDXInputOptions *opts);
	bool VDXAPIENTRY Append(const wchar_t *szFile);

	bool VDXAPIENTRY PromptForOptions(VDXHWND, IVDXInputOptions **);
	bool VDXAPIENTRY CreateOptions(const void *buf, uint32 len, IVDXInputOptions **);
	void VDXAPIENTRY DisplayInfo(VDXHWND hwndParent);

	bool VDXAPIENTRY GetVideoSource(int index, IVDXVideoSource **);
	bool VDXAPIENTRY GetAudioSource(int index, IVDXAudioSource **);

	unsigned long frames;
	__int64 *framelist;

	void ReadSample(void *buffer, unsigned long sample, long len);

protected:
	const VDInputDriverContext& mContext;

private:
	bool guiDlgMessageLoop(HWND hDlg);
	static INT_PTR CALLBACK ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

	inFile mFile;
	bool fAbort;
};


///////////////////////////////////////////////////////////////////////
//
//						AudioSourceAC3
//
///////////////////////////////////////////////////////////////////////

static const short rate[19] = {
	32, 40, 48, 56, 64, 80, 96, 112, 128, 160,
	192, 224, 256, 320, 384, 448, 512, 576, 640
};

// The position of the LFE bit depends on what follows acmod:
// if ((acmod & 1) && (acmod != 1)) cmixlev (2 bits) follows acmod
// if (acmod & 4) surmixlev (2 bits) follows acmod (and cmixlev)

// Bytes 6-7, x = LFE:
//	000x.... ........	1+1, LFE
//  001x.... ........	1/0, LFE
//  010x.... ........	2/0, LFE
//  011..x.. ........	3/0, cmixlev, LFE
//  100..x.. ........	2/1, surmixlev, LFE
//  101....x ........	3/1, cmixlev, surmixlev, LFE
//	110..x.. ........	2/2, surmixlev, LFE
//	111....x ........	3/2, cmixlev, surmixlev, LFE

static const short lfeon[8] = {
	0x1000, 0x1000, 0x1000, 0x0400,	0x0400, 0x0100, 0x0400, 0x0100
};

static const char chan[8] = { 2, 1, 2, 3, 3, 4, 4, 5 };


// bytes 4-7
// xx...... ........ ........ ........  fscod (00=48, 01=44.1, 10=32, 11=reserved)
// ..xxxxxx ........ ........ ........  frmsizecod (0 to 37)
// ........ xxxxx... ........ ........  bsid (01000 = version 8)
// ........ .....xxx ........ ........  bsmod
// ........ ........ xxx..... ........  acmod (#channels)
// ........ ........ ...x.x.x ........  lfe (position depends on acmod)

unsigned short ac3_compact_header(const unsigned long hdr) {
	const int acmod = (hdr >> 13) & 7;

	return (unsigned short)(
		((hdr >> 16) & 0xFF00) |
		(acmod << 1) | !!(hdr & lfeon[acmod])
	);
}

// Compact ac3 header info by eliminating unneeded bits
// xx...... ........  fscod (00=48, 01=44.1, 10=32, 11=reserved)
// ..xxxxxx ........  frmsizecod (0 to 37)
// ........ 0000....  unused
// ........ ....xxx.  acmod (#channels)
// ........ .......x  lfe (derived at compaction time)

int ac3_compact_info(const unsigned short hdr, long *srate, int *brate, int *chans) {
	const int frmsizecod = (hdr >> 8) & 63;
	const int bitrate = rate[frmsizecod >> 1];
	const int acmod = (hdr >> 1) & 7;

// We need not verify anything here, because
// any compacted header has already been verified.

	if (chans) *chans = chan[acmod] + (hdr & 1);
	if (brate) *brate = bitrate;

	switch (hdr & 0xC000) {
		case 0:
			if (srate) *srate = 48000;
			return (4 * bitrate);
		case 0x4000:
			if (srate) *srate = 44100;
			return (2 * (320 * bitrate / 147 + (frmsizecod & 1)));
		case 0x8000:
			if (srate) *srate = 32000;
			return (6 * bitrate);
	}
	return 0;
}

// bytes 4-7
// xx...... ........ ........ ........  fscod (00=48, 01=44.1, 10=32, 11=reserved)
// ..xxxxxx ........ ........ ........  frmsizecod (0 to 37)
// ........ xxxxx... ........ ........  bsid (01000 = version 8)
// ........ .....xxx ........ ........  bsmod
// ........ ........ xxx..... ........  acmod (#channels)
// ........ ........ ...x.x.x ........  lfe (position depends on acmod)

int ac3_sync_info(const unsigned long hdr, long *srate, int *brate, int *chans) {

	if (hdr < 0xC0000000) {
		const int frmsizecod = (hdr >> 24) & 63;

		if (frmsizecod < 38) {
			const int bsid = (hdr >> 19) & 31;

			if (bsid <= 8) {
				const int acmod = (hdr >> 13) & 7;
				const int bitrate = rate[frmsizecod >> 1];

				if (chans) *chans = chan[acmod] + !!(hdr & lfeon[acmod]);
				if (brate) *brate = bitrate;

				switch (hdr >> 30) {
					case 0:
						if (srate) *srate = 48000;
						return (4 * bitrate);
					case 1:
						if (srate) *srate = 44100;
						return (2 * (320 * bitrate / 147 + (frmsizecod & 1)));
					case 2:
						if (srate) *srate = 32000;
						return (6 * bitrate);
				}
			}
		}
	}
	return 0;
}

class AudioSourceAC3 : public vdxunknown<IVDXStreamSource>, public IVDXAudioSource {

public:
	AudioSourceAC3(InputFileAC3 *const pp);
	~AudioSourceAC3();

	int VDXAPIENTRY AddRef();
	int VDXAPIENTRY Release();
	void *VDXAPIENTRY AsInterface(uint32 iid);

	void		VDXAPIENTRY GetStreamSourceInfo(VDXStreamSourceInfo&);
	bool		VDXAPIENTRY Read(sint64 lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);

	const void *VDXAPIENTRY GetDirectFormat();
	int			VDXAPIENTRY GetDirectFormatLen();

	ErrorMode VDXAPIENTRY GetDecodeErrorMode();
	void VDXAPIENTRY SetDecodeErrorMode(ErrorMode mode);
	bool VDXAPIENTRY IsDecodeErrorModeSupported(ErrorMode mode);

	bool VDXAPIENTRY IsVBR();
	sint64 VDXAPIENTRY TimeToPositionVBR(sint64 us);
	sint64 VDXAPIENTRY PositionToTimeVBR(sint64 samples);

	void VDXAPIENTRY GetAudioSourceInfo(VDXAudioSourceInfo& info);

protected:
	InputFileAC3 *const parentPtr;

	struct {
		WAVEFORMATEX wfex;
		WAVEFORMATEX wfexDummy;
	} mRawFormat;

	unsigned long mLength;
};

AudioSourceAC3::AudioSourceAC3(InputFileAC3 *pp)
	: parentPtr(pp)
{
	parentPtr->AddRef();
	mLength = 0;

	if (parentPtr->framelist && parentPtr->frames) {
		int chans, brate;
		long srate;

		int frmsize = ac3_compact_info((unsigned short)
			(parentPtr->framelist[0] & 0xFFFF),
			&srate, &brate, &chans);

		memset(&mRawFormat, 0, sizeof(mRawFormat));

		mRawFormat.wfex.wFormatTag		= 0x2000;
		mRawFormat.wfex.nChannels		= chans;
		mRawFormat.wfex.nSamplesPerSec	= srate;
		mRawFormat.wfex.nAvgBytesPerSec	= brate * 125;
		mRawFormat.wfex.nBlockAlign		= frmsize;
		mRawFormat.wfex.wBitsPerSample	= 0;
		mRawFormat.wfex.cbSize			= sizeof(WAVEFORMATEX);

		mLength = parentPtr->frames;
	}
}

AudioSourceAC3::~AudioSourceAC3() {
	parentPtr->Release();
}

int AudioSourceAC3::AddRef() {
	return vdxunknown<IVDXStreamSource>::AddRef();
}

int AudioSourceAC3::Release() {
	return vdxunknown<IVDXStreamSource>::Release();
}

void *VDXAPIENTRY AudioSourceAC3::AsInterface(uint32 iid) {
	if (iid == IVDXAudioSource::kIID)
		return static_cast<IVDXAudioSource *>(this);

	if (iid == IVDXStreamSource::kIID)
		return static_cast<IVDXStreamSource *>(this);

	return NULL;
}

void VDXAPIENTRY AudioSourceAC3::GetStreamSourceInfo(VDXStreamSourceInfo& srcInfo) {
	if (mLength) {
		srcInfo.mSampleRate.mNumerator = mRawFormat.wfex.nAvgBytesPerSec;
		srcInfo.mSampleRate.mDenominator = mRawFormat.wfex.nBlockAlign;
		srcInfo.mSampleCount = mLength;
	} else {
		srcInfo.mSampleRate.mNumerator = 0;
		srcInfo.mSampleRate.mDenominator = 0;
		srcInfo.mSampleCount = 0;
	}
}

bool AudioSourceAC3::Read(sint64 lStart64, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {

	uint32 samples;
	const unsigned long lStart = (unsigned long)lStart64;
	int frmsize = mRawFormat.wfex.nBlockAlign;

	if (lStart >= mLength) {
		if (lSamplesRead) *lSamplesRead = 0;
		if (lBytesRead) *lBytesRead = 0;
		return true;
	}

	if (lpBuffer == NULL || cbBuffer < (uint32)frmsize) {
		if (lSamplesRead) *lSamplesRead = 1;
		if (lBytesRead) *lBytesRead = frmsize;
		return (lpBuffer == NULL);
	}

	samples = cbBuffer / frmsize;
	if (lCount > samples) lCount = samples;

	// Don't read past EOF
	if ((lStart + lCount) > mLength) {
		lCount = (uint32)(mLength - lStart);
	}

	if (lSamplesRead) *lSamplesRead = lCount;
	lCount *= frmsize;
	parentPtr->ReadSample(lpBuffer, lStart, lCount);

	if (lBytesRead) *lBytesRead = lCount;
	return true;
}

const void *AudioSourceAC3::GetDirectFormat() {
	if (mLength) return &mRawFormat;
	return NULL;
}

int AudioSourceAC3::GetDirectFormatLen() {
	if (mLength) return sizeof(mRawFormat);
	return 0;
}

IVDXStreamSource::ErrorMode AudioSourceAC3::GetDecodeErrorMode() {
	return IVDXStreamSource::kErrorModeReportAll;
}

void AudioSourceAC3::SetDecodeErrorMode(IVDXStreamSource::ErrorMode mode) {
}

bool AudioSourceAC3::IsDecodeErrorModeSupported(IVDXStreamSource::ErrorMode mode) {
	return mode == IVDXStreamSource::kErrorModeReportAll;
}

bool AudioSourceAC3::IsVBR() {
	return false;
}

sint64 AudioSourceAC3::TimeToPositionVBR(sint64 us) {
	return 0;
}

sint64 AudioSourceAC3::PositionToTimeVBR(sint64 samples) {
	return 0;
}

void AudioSourceAC3::GetAudioSourceInfo(VDXAudioSourceInfo& info) {
	info.mFlags = 0;
}


///////////////////////////////////////////////////////////////////////
//
//						DataVector
//
///////////////////////////////////////////////////////////////////////

class DataVector {

private:
	struct DataVectorBlock {
		enum { BLOCK_SIZE = 4096-12 };

		DataVectorBlock *prev;
		DataVectorBlock *next;
		int size;
		char heap[BLOCK_SIZE];
	};

	DataVectorBlock *first, *last;
	const int _item_size;
	unsigned long count;

public:
	DataVector(const int item_size);
	~DataVector();

	void Add(const void *pp);
	void *MakeArray();

	unsigned long Length() { return count; };
};

DataVector::DataVector(int item_size)
	: _item_size(item_size)
{
	first = last = NULL;
	count = 0;
}

DataVector::~DataVector() {
	DataVectorBlock *i, *j;

	i = last;
	while (j = i) {
		i = j->prev;
		delete j;
	}
}

void DataVector::Add(const void *pp) {

	int len = _item_size;

	if (last == NULL) {
		DataVectorBlock *ib = new DataVectorBlock;

		memcpy(ib->heap, pp, len);
		ib->next = NULL;
		ib->size = len;

		if (first == NULL) {
			// New "first" block
			first = ib;
			first->prev = NULL;
		} else {
			// New "last" block
			last = ib;
			first->next = last;
			last->prev = first;
		}

	} else {
		const int remains = DataVectorBlock::BLOCK_SIZE - last->size;

		if (len > remains) {
			// New "last" block
			DataVectorBlock *ib = new DataVectorBlock;

			last->next = ib;
			ib->prev = last;
			ib->next = NULL;
			ib->size = 0;

			if (remains > 0) {
				memcpy(last->heap + last->size, pp, remains);
				pp = (char *)pp + remains;
				last->size += remains;
				len -= remains;
			}

			last = ib;
		}

		memcpy(last->heap + last->size, pp, len);
		last->size += len;
	}

	++count;
}

void *DataVector::MakeArray() {

	char *array = new char[count * _item_size];
	char *ptr = array;

	const DataVectorBlock *dvb = first;

	while (dvb) {
		memcpy(ptr, dvb->heap, dvb->size);
		ptr += dvb->size;

		dvb = dvb->next;
	}

	return array;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileAC3
//
///////////////////////////////////////////////////////////////////////

InputFileAC3::InputFileAC3(const VDInputDriverContext& context)
	: mContext(context)
{
	framelist = NULL;
	fAbort = false;
	frames = 0;
}

INT_PTR CALLBACK InputFileAC3::ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	InputFileAC3 *thisPtr = (InputFileAC3 *)GetWindowLongPtr(hDlg, DWLP_USER);

	switch(uMsg) {

	case WM_INITDIALOG:
		{
			SetWindowLongPtr(hDlg, DWLP_USER, lParam);
			thisPtr = (InputFileAC3 *)lParam;

			SendMessage(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETRANGE, 0, MAKELPARAM(0, 16384));
			SetTimer(hDlg, 1, 250, (TIMERPROC)NULL);
			ShowWindow(hDlg, SW_SHOW);

			return TRUE;
		}

	case WM_TIMER:
		{
			char buf[64];
			__int64 file_cpos = thisPtr->mFile.inPos();
			__int64 file_size = thisPtr->mFile.inSize();

			SendMessage(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETPOS,
				(WPARAM)((file_cpos * 16384I64) / file_size), 0);

			sprintf(buf, "%ldK of %ldK",
				(long)(file_cpos >> 10),
				(long)((file_size + 512) >> 10));

			SendDlgItemMessage(hDlg, IDC_CURRENT_VALUE, WM_SETTEXT, 0, (LPARAM)buf);
		}
		return TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDCANCEL) thisPtr->fAbort = true;

		return TRUE;
	}

	return FALSE;
}

bool InputFileAC3::guiDlgMessageLoop(HWND hDlg) {
	MSG msg;

	while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
		if (msg.message == WM_QUIT) return false;

		if (!hDlg || !IsWindow(hDlg) || !IsDialogMessage(hDlg, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return true;
}

void InputFileAC3::Init(const wchar_t *szFile, IVDXInputOptions *opts) {
	DataVector dv(sizeof(__int64));
	unsigned short syncword;
	unsigned long header;
	int first, bytes, limit;
	HWND hwndProgress;

	if (!mFile.inOpen(szFile)) {
		mContext.mpCallbacks->SetError("Can't open %ls", szFile);
		return;
	}

	syncword = 0;
	header = 0;
	first = 0;
	bytes = 0;
	limit = 0;

	// Pop up parse dialog
	hwndProgress = CreateDialogParam(g_hModule,
		MAKEINTRESOURCE(IDD_PROGRESS), GetActiveWindow(),
		ParseDialogProc, (LPARAM)this);

	do {
		
		if (syncword == 0x0B77) {
			// Got syncword, now get the 6 bytes which follow it

			if (bytes == 6) {
				int frmsize = ac3_sync_info(header, NULL, NULL, NULL);

				if (frmsize >= 8) {

					// If the final frame is incomplete, don't add it
					if ((mFile.inPos() + frmsize) > mFile.inSize()) break;

					// We can't support VBR yet
					if (frmsize != first) {
						if (first != 0) {
							mContext.mpCallbacks->SetError("Variable bitrate AC-3 is not supported.");
							break;
						}
						first = frmsize;
					}

					__int64 info = ((mFile.inPos() - 8) << 16)
						+ ac3_compact_header(header);

					dv.Add(&info);
					++frames;

					// Update the progress dialog
					guiDlgMessageLoop(hwndProgress);
					if (fAbort) break;

					// The smallest possible AC-3 frame is 192 bytes
					if ((mFile.inSize() - mFile.inPos()) < 192) break;

					mFile.inSeek(mFile.inPos() + (frmsize - 8));

					limit = 0;	// reset sync hunt
				}

				// Go back to hunting for syncword
				syncword = 0;
				bytes = 0;

			} else {

				// Accumulate header bytes
				header = (header << 8) + mFile.inByte();
				++bytes;
			}

		} else {

			// Hunt 4K for 0x0B77 syncword
			syncword = (syncword << 8) + mFile.inByte();
			++limit;
		}

	} while (limit < 4096);

	if (hwndProgress != NULL) {
		DestroyWindow(hwndProgress);
	}

	if (frames > 0) {
		framelist = (__int64 *)(dv.MakeArray());
		if (framelist == NULL) {
			frames = 0;
			mContext.mpCallbacks->SetErrorOutOfMemory();
		}
	} else {
		mContext.mpCallbacks->SetError("Invalid AC-3 stream.");
	}
}

bool InputFileAC3::Append(const wchar_t *szFile) {
	return false;
}

bool InputFileAC3::PromptForOptions(VDXHWND, IVDXInputOptions **ppOptions) {
	*ppOptions = NULL;
	return false;
}

bool InputFileAC3::CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions) {
	*ppOptions = NULL;
	return false;
}

void InputFileAC3::DisplayInfo(VDXHWND hwndParent) {
}

bool InputFileAC3::GetVideoSource(int index, IVDXVideoSource **ppVS) {
	*ppVS = NULL;
	return false;
}

bool InputFileAC3::GetAudioSource(int index, IVDXAudioSource **ppAS) {
	*ppAS = NULL;

	if (index) return false;

	IVDXAudioSource *pAS = new AudioSourceAC3(this);
	if (!pAS) return false;

	*ppAS = pAS;
	pAS->AddRef();
	return true;
}

void InputFileAC3::ReadSample(void *buffer, unsigned long sample, long len) {
	__int64 offset;
	int frmsize;

	while (len > 0) {
		if (sample >= frames) break;

		offset = framelist[sample] >> 16;
		frmsize = ac3_compact_info((unsigned short)
			(framelist[sample] & 0xFFFF), NULL, NULL, NULL);

		if ((offset + frmsize) >= mFile.inSize()) {
			mContext.mpCallbacks->SetError("Attempt to read past end of audio");
			break;
		}

		long tc = frmsize;
		if (tc > len) tc = len;

		mFile.inSeek(offset);
		mFile.inRead(buffer, tc);

		buffer = (char *)buffer + tc;
		len -= tc;
		++sample;
	}
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileDriverAC3
//
///////////////////////////////////////////////////////////////////////

class InputFileDriverAC3 : public vdxunknown<IVDXInputFileDriver> {
public:
	InputFileDriverAC3(const VDInputDriverContext& context);

	int		VDXAPIENTRY DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize);
	bool	VDXAPIENTRY CreateInputFile(uint32 flags, IVDXInputFile **ppFile);

protected:
	int		mRefCount;

	const VDInputDriverContext& mContext;
};

InputFileDriverAC3::InputFileDriverAC3(const VDInputDriverContext& context)
	: mRefCount(0)
	, mContext(context)
{
}

int VDXAPIENTRY InputFileDriverAC3::DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize) {
	return -1;
}

bool VDXAPIENTRY InputFileDriverAC3::CreateInputFile(uint32 flags, IVDXInputFile **ppFile) {
	IVDXInputFile *p = new InputFileAC3(mContext);
	if (!p)
		return false;

	*ppFile = p;
	p->AddRef();
	return true;
}

bool VDXAPIENTRY ac3_create(const VDInputDriverContext *pContext, IVDXInputFileDriver **ppDriver) {
	IVDXInputFileDriver *p = new InputFileDriverAC3(*pContext);
	if (!p) return false;

	*ppDriver = p;
	p->AddRef();
	return true;
}

const uint8 ac3_sig[] = {
	0x0B, 0xFF, 0x77, 0xFF,		// syncword
	0, 0, 0, 0,					// crc1
	0, 0,						// fscod, frmsizecod
	0x40, 0xF8					// bsid = 8
};

const VDInputDriverDefinition ac3_input = {
	sizeof(VDInputDriverDefinition),
	VDInputDriverDefinition::kFlagSupportsAudio,
	0,
	sizeof ac3_sig,
	ac3_sig,
	L"*.ac3",
	L"Dolby Digital (*.ac3)|*.ac3",
	L"ac3",
	ac3_create
};

const VDPluginInfo ac3_info = {
	sizeof(VDPluginInfo),
	L"AC3 plugin",
	L"fccHandler",
	L"Loads Dolby Digital audio.",
	0x01020000,
	kVDPluginType_Input,
	0,
	10,
	kVDPlugin_APIVersion,
	2,
	kVDPlugin_InputDriverAPIVersion,
	&ac3_input
};

///////////////////////////////////////////////////////////////////////

const VDPluginInfo *const kPlugins[] = {
	&ac3_info,
	NULL
};

extern "C" __declspec(dllexport) const VDPluginInfo *const * VDXAPIENTRY VDGetPluginInfo() {
	return kPlugins;
}
